"""
Exceptions specific to the Google Sheets adapter.
"""

from shillelagh.exceptions import Error


class DateParseError(Error):
    """
    Raised when a date/time value cannot be parsed from a given pattern.
    """



"""
An adapter for fetching XML data.
"""

import logging
import xml.etree.ElementTree as ET
from collections.abc import Iterator
from typing import Any, Optional

from defusedxml import ElementTree as DET

from shillelagh.adapters.api.generic_json import GenericJSONAPI
from shillelagh.exceptions import ProgrammingError
from shillelagh.filters import Filter
from shillelagh.lib import flatten
from shillelagh.typing import RequestedOrder, Row

_logger = logging.getLogger(__name__)


def element_to_dict(element: ET.Element) -> Any:
    """
    Convert XML element to a dictionary, recursively.

    This uses a super simple algorithm that focuses on text and ignores attributes.
    """
    if element.text and element.text.strip():
        return element.text.strip()

    result: dict[str, Any] = {}
    for child in element:
        child_data = element_to_dict(child)
        if child.tag in result:
            # Convert to a list if multiple elements with the same tag exist
            if not isinstance(result[child.tag], list):
                result[child.tag] = [result[child.tag], child_data]
            else:
                result[child.tag].append(child_data)
        else:
            result.update({child.tag: child_data})

    return result


class GenericXMLAPI(GenericJSONAPI):
    """
    An adapter for fetching XML data.
    """

    safe = True

    supports_limit = False
    supports_offset = False
    supports_requested_columns = True

    content_type = "xml"  # works with text/xml and application/xml
    default_path = "*"
    cache_name = "generic_xml_cache"

    def get_data(  # pylint: disable=unused-argument, too-many-arguments
        self,
        bounds: dict[str, Filter],
        order: list[tuple[str, RequestedOrder]],
        requested_columns: Optional[set[str]] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        response = self._session.get(self.uri)
        payload = response.content.decode("utf-8")
        if not response.ok:
            raise ProgrammingError(f"Error: {payload}")

        root = DET.fromstring(payload)
        result = root.findall(self.path)
        for i, element in enumerate(result):
            row = element_to_dict(element)
            row = {
                k: v
                for k, v in row.items()
                if requested_columns is None or k in requested_columns
            }
            row["rowid"] = i
            _logger.debug(row)
            yield flatten(row)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=unused-argument, abstract-method

"""
A SQLAlchemy dialect for dbt Metric Flow.
"""

from __future__ import annotations

from typing import Any, cast

import sqlalchemy.types
from sqlalchemy.engine.base import Connection
from sqlalchemy.engine.url import URL
from sqlalchemy.sql.type_api import TypeEngine

from shillelagh.adapters.api.dbt_metricflow import DbtMetricFlowAPI
from shillelagh.adapters.registry import registry
from shillelagh.backends.apsw.dialects.base import (
    APSWDialect,
    SQLAlchemyColumn,
    get_adapter_for_table_name,
)
from shillelagh.fields import Field

TABLE_NAME = "metrics"


def get_sqla_type(field: Field) -> type[TypeEngine]:
    """
    Convert from Shillelagh to SQLAlchemy types.
    """
    type_map: dict[str, type[TypeEngine]] = {
        "BOOLEAN": sqlalchemy.types.BOOLEAN,
        "INTEGER": sqlalchemy.types.INT,
        "DECIMAL": sqlalchemy.types.DECIMAL,
        "TIMESTAMP": sqlalchemy.types.TIMESTAMP,
        "DATE": sqlalchemy.types.DATE,
        "TIME": sqlalchemy.types.TIME,
        "TEXT": sqlalchemy.types.TEXT,
    }

    return type_map.get(field.type, sqlalchemy.types.TEXT)


class TableMetricFlowAPI(DbtMetricFlowAPI):
    """
    Custom API adapter for dbt Metric Flow API.

    In the original adapter, the SQL queries a base dbt API URL, eg:

        SELECT * FROM "https://semantic-layer.cloud.getdbt.com/";
        SELECT * FROM "https://ab123.us1.dbt.com/";  -- custom user URL

    For this adapter, we want a leaner URI, mimicking a table:

        SELECT * FROM metrics;

    In order to do this, we override the ``supports`` method to only accept
    ``$TABLE_NAME`` instead of the URL, which is then passed to the adapter when it is
    instantiated.
    """

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> bool:
        return uri == TABLE_NAME

    def __init__(
        self,
        table: str,
        service_token: str,
        environment_id: int,
        url: str,
    ) -> None:
        # pass URL as the table name to the base class, since that's the actual URL
        super().__init__(url, service_token, environment_id)

        # but make sure we set the table name to the expected one, since `get_data`
        # needs it
        self.table = table


class MetricFlowDialect(APSWDialect):
    """
    A dbt Metric Flow dialect.

    URL should look like:

        metricflow:///<environment_id>?service_token=<service_token>

    Or when using a custom URL:

        metricflow://ab123.us1.dbt.com/<environment_id>?service_token=<service_token>

    """

    name = "metricflow"

    supports_statement_cache = True

    def __init__(
        self,
        service_token: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        self.service_token = service_token

    def create_connect_args(self, url: URL) -> tuple[tuple[()], dict[str, Any]]:
        baseurl = (
            f"https://{url.host}/"
            if url.host
            else "https://semantic-layer.cloud.getdbt.com/"
        )

        if "tablemetricflowapi" not in registry.loaders:
            registry.add("tablemetricflowapi", TableMetricFlowAPI)

        return (
            (),
            {
                "path": ":memory:",
                "adapters": ["tablemetricflowapi"],
                "adapter_kwargs": {
                    "tablemetricflowapi": {
                        "service_token": url.query.get(
                            "service_token",
                            self.service_token,
                        ),
                        "environment_id": int(url.database),
                        "url": baseurl,
                    },
                },
                "safe": True,
                "isolation_level": self.isolation_level,
            },
        )

    def get_table_names(
        self,
        connection: Connection,
        schema: str | None = None,
        sqlite_include_internal: bool = False,
        **kwargs: Any,
    ) -> list[str]:
        return [TABLE_NAME]

    def has_table(
        self,
        connection: Connection,
        table_name: str,
        schema: str | None = None,
        **kwargs: Any,
    ) -> bool:
        return table_name == TABLE_NAME

    def get_columns(
        self,
        connection: Connection,
        table_name: str,
        schema: str | None = None,
        **kwargs: Any,
    ) -> list[SQLAlchemyColumn]:
        columns: list[SQLAlchemyColumn] = []
        adapter = cast(
            DbtMetricFlowAPI,
            get_adapter_for_table_name(connection, table_name),
        )

        dimensions = {
            (
                adapter.grains[dimension][0]
                if dimension in adapter.grains
                else dimension
            ): adapter.columns[dimension]
            for dimension in adapter.dimensions
        }
        for name, field in dimensions.items():
            columns.append(
                {
                    "name": name,
                    "type": get_sqla_type(field),
                    "nullable": True,
                    "default": None,
                    "comment": adapter.dimensions.get(name, ""),
                },
            )

        metrics = {metric: adapter.columns[metric] for metric in adapter.metrics}
        for name, field in metrics.items():
            columns.append(
                {
                    "name": name,
                    "type": get_sqla_type(field),
                    "nullable": True,
                    "default": None,
                    "comment": adapter.metrics.get(name, ""),
                    "computed": {"sqltext": name, "persisted": True},
                },
            )

        return columns

    def get_schema_names(
        self,
        connection: Connection,
        **kwargs: Any,
    ) -> list[str]:
        return ["main"]

    def get_pk_constraint(
        self,
        connection: Connection,
        table_name: str,
        schema: str | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        return {"constrained_columns": [], "name": None}

    def get_foreign_keys(
        self,
        connection: Connection,
        table_name: str,
        schema: str | None = None,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        return []

    get_check_constraints = get_foreign_keys
    get_indexes = get_foreign_keys
    get_unique_constraints = get_foreign_keys

    def get_table_comment(self, connection, table_name, schema=None, **kwargs):
        return {
            "text": "A virtual table that gives access to all dbt metrics & dimensions.",
        }



# pylint: disable=invalid-name, c-extension-no-member, unused-import
"""
A DB API 2.0 wrapper based on sqlglot.
"""

import datetime
import logging
from collections import defaultdict
from typing import Any, DefaultDict, Optional

import sqlglot
from sqlglot import exp
from sqlglot.executor import execute
from sqlglot.optimizer.annotate_types import annotate_types
from sqlglot.optimizer.pushdown_predicates import pushdown_predicates
from sqlglot.optimizer.qualify_columns import qualify_columns
from sqlglot.optimizer.scope import traverse_scope
from sqlglot.schema import MappingSchema

from shillelagh.adapters.base import Adapter
from shillelagh.adapters.registry import registry
from shillelagh.backends.apsw.db import DEFAULT_SCHEMA
from shillelagh.db import (
    Connection,
    Cursor,
    apilevel,
    check_closed,
    paramstyle,
    threadsafety,
)
from shillelagh.exceptions import (  # nopycln: import; pylint: disable=redefined-builtin
    DatabaseError,
    DataError,
    Error,
    IntegrityError,
    InterfaceError,
    InternalError,
    OperationalError,
    ProgrammingError,
    Warning,
)
from shillelagh.fields import Boolean, DateTime, Field, Integer, String
from shillelagh.filters import Operator
from shillelagh.lib import find_adapter, get_bounds
from shillelagh.types import (
    BINARY,
    DATETIME,
    NUMBER,
    ROWID,
    STRING,
    Binary,
    Date,
    DateFromTicks,
    Time,
    TimeFromTicks,
    Timestamp,
    TimestampFromTicks,
)

__all__ = [
    "DatabaseError",
    "DataError",
    "Error",
    "IntegrityError",
    "InterfaceError",
    "InternalError",
    "OperationalError",
    "BINARY",
    "DATETIME",
    "NUMBER",
    "ROWID",
    "STRING",
    "Binary",
    "Date",
    "DateFromTicks",
    "Time",
    "TimeFromTicks",
    "Timestamp",
    "TimestampFromTicks",
    "Warning",
    "apilevel",
    "threadsafety",
    "paramstyle",
]


_logger = logging.getLogger(__name__)


type_map: dict[exp.DataType.Type, type[Field]] = {
    exp.DataType.Type.INT: Integer,
    exp.DataType.Type.TEXT: String,
    exp.DataType.Type.BOOLEAN: Boolean,
    exp.DataType.Type.TIMESTAMP: DateTime,
}

DEFAULT_TYPE = String


def remove_anded_parentheses(node: exp.Expression) -> exp.Expression:
    """
    Remove unnecessary parentheses around `AND` expressions.
    """
    return (
        node.this.transform(remove_anded_parentheses)
        if isinstance(node, exp.Paren) and isinstance(node.this, exp.And)
        else node
    )


def to_py(ast: exp.Expression) -> Any:
    """
    Convert an expression to a Python value.
    """
    if isinstance(ast, exp.TimeStrToTime):
        return datetime.datetime.fromisoformat(ast.this.to_py())

    if isinstance(ast, exp.DateStrToDate):
        return datetime.date.fromisoformat(ast.this.to_py())

    if (
        # remove once https://github.com/tobymao/sqlglot/pull/5409 is released
        hasattr(exp, "DateStrToTime") and isinstance(ast, exp.DateStrToTime)  # pylint: disable=no-member
    ):
        return datetime.time.fromisoformat(ast.this.to_py())  # pragma: no cover

    return ast.to_py()


class SQLGlotCursor(Cursor):  # pylint: disable=too-many-instance-attributes
    """
    Connection cursor.
    """

    @check_closed
    def execute(
        self,
        operation: str,
        parameters: Optional[tuple[Any, ...]] = None,
    ) -> "SQLGlotCursor":
        """
        Execute a query using the cursor.
        """
        self.description = None
        self._rowcount = -1

        # store current SQL in the cursor
        self.operation = operation
        try:
            ast = sqlglot.parse_one(operation, "sqlite")
        except sqlglot.errors.ParseError as exc:
            raise ProgrammingError("Invalid SQL query") from exc

        # drop table?
        if uri := self._drop_table_uri(ast):
            adapter = self._get_adapter_instance(uri)
            adapter.drop_table()
            return self

        if not isinstance(ast, exp.Select):
            raise InterfaceError("Only `DROP TABLE` and `SELECT` queries are supported")

        # run query
        if parameters:
            ast = exp.replace_placeholders(ast, *parameters)

        # qualify query so we can push down predicates to adapters
        schema = self._get_schema(ast)
        qualified = qualify_columns(ast, schema=schema)
        annotated = annotate_types(qualified, schema=schema)
        tables = self._get_tables(annotated)

        # and execute query
        table = execute(annotated, schema=schema, tables=tables)
        self._results = (reader.row for reader in table)

        # store description
        self.description = [
            (
                name,
                type_map.get(expression.type.this, DEFAULT_TYPE),
                None,
                None,
                None,
                None,
                True,
            )
            for name, expression in zip(table.columns, annotated.expressions)
        ]

        return self

    def _get_adapter_instance(self, uri: str) -> Adapter:
        """
        Return an adapter instance for the given URI.
        """
        adapter, args, kwargs = find_adapter(uri, self._adapter_kwargs, self._adapters)

        return adapter(*args, **kwargs)

    def _drop_table_uri(self, ast: exp.Expression) -> Optional[str]:
        """
        Extract table being dropped, if any
        """
        if not isinstance(ast, exp.Drop):
            return None

        return ast.find(exp.Table).name if ast.find(exp.Table) else None

    def _get_schema(self, ast: exp.Select) -> MappingSchema:
        """
        Return the schema of all referenced tables.
        """
        schema = MappingSchema(dialect="sqlite")

        for relation in self._get_relations(ast):
            uri = relation.name
            adapter = self._get_adapter_instance(uri)
            column_mapping = {
                name: field.type for name, field in adapter.get_columns().items()
            }
            schema.add_table(
                table=exp.Table(this=exp.Identifier(this=uri, quoted=True)),
                column_mapping=column_mapping,
            )

        return schema

    def _get_relations(self, ast: exp.Select) -> set[exp.Table]:
        """
        Return the relations referenced in the query.
        """
        return {
            source
            for scope in traverse_scope(ast)
            for source in scope.sources.values()
            if isinstance(source, exp.Table)
        }

    def _get_tables(self, ast: exp.Select) -> dict[str, list[dict[str, Any]]]:
        """
        Build the tables needed for the sqlglot executor.
        """
        modified = ast.copy()

        # first replace all table names with a dummy subquery, so we can push predicates
        # more easily
        for relation in self._get_relations(modified):
            identifier = exp.Identifier(this=relation.name, quoted=True)
            relation.replace(
                exp.Subquery(
                    this=exp.Select(
                        expressions=[exp.Star()],
                        **{"from": exp.From(this=exp.Table(this=identifier))},
                    ),
                    alias=exp.TableAlias(this=identifier),
                ),
            )

        # now push predicates
        modified = pushdown_predicates(modified)

        # finally build a table for each subquery
        tables = {}
        for subquery in modified.find_all(exp.Subquery):
            if not isinstance(subquery.this.args["from"].this, exp.Table):
                continue

            uri = subquery.alias
            adapter = self._get_adapter_instance(uri)
            columns = adapter.get_columns()
            all_bounds = self._get_all_bounds(columns, subquery)
            bounds = get_bounds(columns, all_bounds)
            table = exp.Table(this=exp.Identifier(this=uri, quoted=True))
            tables[table] = list(adapter.get_rows(bounds, order=[]))

        return tables

    def _get_all_bounds(  # pylint: disable=too-many-branches
        self,
        columns: dict[str, Field],
        ast: exp.Subquery,
    ) -> DefaultDict[str, set[tuple[Operator, Any]]]:
        """
        Convert predicates to bounds whenever possible.
        """
        all_bounds: DefaultDict[str, set[tuple[Operator, Any]]] = defaultdict(set)

        if "where" not in ast.this.args:
            return all_bounds

        where = ast.this.args["where"]
        where = where.transform(remove_anded_parentheses)

        predicates: list[exp.Expression] = []
        predicate = where.this
        while isinstance(predicate, exp.And):
            predicates.append(predicate.expression)
            predicate = predicate.this
        predicates.append(predicate)

        for predicate in predicates:
            if self._is_valid_column_predicate(predicate) and isinstance(
                predicate,
                (exp.GTE, exp.GT, exp.LTE, exp.LT, exp.EQ, exp.NEQ, exp.Like),
            ):
                operator = {
                    exp.GTE: Operator.GE,
                    exp.GT: Operator.GT,
                    exp.LTE: Operator.LE,
                    exp.LT: Operator.LT,
                    exp.EQ: Operator.EQ,
                    exp.NEQ: Operator.NE,
                    exp.Like: Operator.LIKE,
                }[type(predicate)]
                all_bounds[predicate.this.name].add(
                    (operator, to_py(predicate.expression)),
                )
            elif isinstance(predicate, exp.Column):
                all_bounds[predicate.name].add((Operator.EQ, True))
            elif isinstance(predicate, exp.Is) and predicate.expression == exp.Null():
                all_bounds[predicate.this.name].add((Operator.IS_NULL, True))
            elif (
                isinstance(predicate, exp.Not)
                and isinstance(predicate.this, exp.Is)
                and predicate.this.expression == exp.Null()
            ):
                all_bounds[predicate.this.this.name].add((Operator.IS_NOT_NULL, True))

        # convert values to types expected by the adapter
        for column_name, operators in all_bounds.items():
            column_type = columns[column_name]
            all_bounds[column_name] = {
                (operator, column_type.format(value)) for operator, value in operators
            }

        return all_bounds

    def _is_valid_column_predicate(self, predicate: exp.Expression) -> bool:
        # sqlglot moves the column to the left side of the operator
        return isinstance(predicate.this, exp.Column) and not isinstance(
            predicate.expression,
            exp.Column,
        )


class SQLGlotConnection(Connection[SQLGlotCursor]):
    """Connection."""

    @check_closed
    def cursor(self) -> SQLGlotCursor:
        """Return a new Cursor Object using the connection."""
        cursor = SQLGlotCursor(self._adapters, self._adapter_kwargs, self.schema)
        self.cursors.append(cursor)

        return cursor


def connect(  # pylint: disable=too-many-arguments, too-many-positional-arguments
    adapters: Optional[list[str]] = None,
    adapter_kwargs: Optional[dict[str, dict[str, Any]]] = None,
    safe: bool = False,
    schema: str = DEFAULT_SCHEMA,
) -> SQLGlotConnection:
    """
    Constructor for creating a connection to the database.
    """
    adapter_kwargs = adapter_kwargs or {}
    enabled_adapters = registry.load_all(adapters, safe)

    # replace entry point names with class names
    mapping = {
        name: adapter.__name__.lower() for name, adapter in enabled_adapters.items()
    }
    adapter_kwargs = {mapping[k]: v for k, v in adapter_kwargs.items() if k in mapping}

    return SQLGlotConnection(
        list(enabled_adapters.values()),
        adapter_kwargs,
        schema,
        safe,
    )



"""
Custom functions available to the SQL backend.
"""

import importlib
import json
import subprocess
import sys
import time
from datetime import datetime, timedelta
from typing import Any, Optional

from packaging.version import InvalidVersion, Version

import shillelagh
from shillelagh.adapters.base import Adapter
from shillelagh.fields import FastISODateTime
from shillelagh.lib import find_adapter

if sys.version_info < (3, 10):
    from importlib_metadata import distribution
else:
    from importlib.metadata import distribution

__all__ = ["upgrade", "sleep", "get_metadata", "version", "date_trunc"]


def upgrade(target_version: str) -> str:
    """
    Upgrade the library to a given version.

    This function attempts to upgrade shillelagh using available package managers.
    It tries multiple package managers in order: uv, pip, pipx.
    """
    try:
        # Validate version format first
        Version(target_version)
    except InvalidVersion:
        return f"Invalid version: {target_version}"

    package_spec = f"shillelagh=={target_version}"

    # List of package managers to try, in order of preference
    package_managers = [
        # uv is fast and modern
        (["uv", "pip", "install", package_spec], "uv"),
        # Standard pip
        ([sys.executable, "-m", "pip", "install", package_spec], "pip"),
        # pipx for isolated environments
        (
            [
                "pipx",
                "upgrade",
                "--pip-args",
                f"shillelagh=={target_version}",
                "shillelagh",
            ],
            "pipx",
        ),
    ]

    errors = []
    for command, manager_name in package_managers:
        try:
            # Try to run the package manager command
            subprocess.run(
                command,
                capture_output=True,
                text=True,
                check=True,
                timeout=300,  # 5 minute timeout
            )
            # If successful, reload the module
            importlib.reload(shillelagh)
            return f"Upgrade to {target_version} successful using {manager_name}."
        except subprocess.CalledProcessError as e:
            errors.append(f"{manager_name}: {e.stderr or e.stdout or 'Command failed'}")
        except FileNotFoundError:
            errors.append(f"{manager_name}: Not found")
        except subprocess.TimeoutExpired:
            errors.append(f"{manager_name}: Timeout after 5 minutes")
        except Exception as e:  # pylint: disable=broad-except
            errors.append(f"{manager_name}: {str(e)}")

    # If all package managers failed, return error message
    error_details = "; ".join(errors)
    return f"Upgrade failed. Tried: {error_details}"


def sleep(seconds: int) -> None:
    """
    Sleep for ``n`` seconds.

    This is useful for troubleshooting timeouts::

        sql> SELECT sleep(60);

    """
    time.sleep(seconds)


def get_metadata(
    adapter_kwargs: dict[str, dict[str, Any]],
    adapters: list[type[Adapter]],
    uri: str,
) -> str:
    """
    Return metadata about a given table.

    Returns the name of the adapter that supports the table, as well
    as any extra metadata provided by the adapter::

        sql> SELECT GET_METADATA("https://docs.google.com/spreadsheets/d/1/edit#gid=0");
        GET_METADATA("https://docs.google.com/spreadsheets/d/1/edit#gid=0")
        -------------------------------------------------------------------
        {
            "extra": {
                "Spreadsheet title": "A title",
                "Sheet title": "Another title"
            },
            "adapter": "GSheetsAPI"
        }

    """
    adapter, args, kwargs = find_adapter(uri, adapter_kwargs, adapters)
    instance = adapter(*args, **kwargs)

    return json.dumps(
        {
            "extra": instance.get_metadata(),
            "adapter": adapter.__name__,
        },
    )


def version() -> str:
    """
    Return the current version of Shillelagh.

    As an example::

        sql> SELECT VERSION();
        VERSION()
        -----------
        0.7.4

    """
    return str(distribution("shillelagh").version)


def date_trunc(  # pylint: disable=too-many-return-statements
    value: Optional[str],
    unit: str,
) -> Optional[str]:
    """
    Truncate a datetime to a given unit.
    """
    field = FastISODateTime()
    timestamp = field.parse(value)
    if timestamp is None:
        return None

    unit = unit.lower()
    if unit == "year":
        truncated = datetime(year=timestamp.year, month=1, day=1)
    elif unit == "quarter":
        month = ((timestamp.month - 1) // 3) * 3 + 1
        truncated = datetime(year=timestamp.year, month=month, day=1)
    elif unit == "month":
        truncated = datetime(year=timestamp.year, month=timestamp.month, day=1)
    elif unit == "week":
        # assumes the week starts on Monday
        start_of_week = timestamp - timedelta(days=timestamp.weekday())
        truncated = datetime(
            year=start_of_week.year,
            month=start_of_week.month,
            day=start_of_week.day,
        )
    elif unit == "day":
        truncated = datetime(
            year=timestamp.year,
            month=timestamp.month,
            day=timestamp.day,
        )
    elif unit == "hour":
        truncated = datetime(
            year=timestamp.year,
            month=timestamp.month,
            day=timestamp.day,
            hour=timestamp.hour,
        )
    elif unit == "minute":
        truncated = datetime(
            year=timestamp.year,
            month=timestamp.month,
            day=timestamp.day,
            hour=timestamp.hour,
            minute=timestamp.minute,
        )
    elif unit == "second":
        truncated = datetime(
            year=timestamp.year,
            month=timestamp.month,
            day=timestamp.day,
            hour=timestamp.hour,
            minute=timestamp.minute,
            second=timestamp.second,
        )
    else:
        raise ValueError(f"Unsupported truncation unit: {unit}")

    return field.format(truncated)



"""Base class for adapters."""

import atexit
import inspect
from collections.abc import Iterator
from typing import Any, Optional

from shillelagh.exceptions import NotSupportedError
from shillelagh.fields import Field, RowID
from shillelagh.filters import Filter, Operator
from shillelagh.typing import RequestedOrder, Row

FIXED_COST = 666


class Adapter:
    """
    An adapter to a table.

    Adapters provide an interface to resources, so they can be queried via
    SQL. An adapter instance represents a virtual table, and the adapter is
    responsible for fetching data and metadata from the resource, and
    possibly insert, delete, or update rows.

    In order to find an adapter responsible for a given table name, adapters
    need to be registered under the "shillelagh.adapter" entry point, eg::

        # setup.cfg
        [options.entry_points]
        shillelagh.adapter =
            custom_adapter = shillelagh.adapters.api.custom:CustomAdapter

    Adapters also need to implement the ``supports`` method. Given a table
    name, the method should return true if the table is supported by the
    adapter.
    """

    # An adapter is considered "safe" when it has no explicit access to the
    # local filesystem. Users can then use the ``shillelagh+safe://`` URI
    # in SQLAlchemy to load only safe adapters, as well as only adapters
    # explicitly listed:
    #
    #     >>> engine = create_engine("shillelagh+safe://", adapters=["gsheetsapi"])
    #
    safe = False

    # if true, a corresponding argument will be passed in the kwargs of
    # ``get_rows`` and ``get_data``
    supports_limit = False
    supports_offset = False

    # if true, the requested columns will be passed to ``get_rows`` and ``get_data``
    supports_requested_columns = False

    def __init__(self, *args: Any, **kwargs: Any):  # pylint: disable=unused-argument
        # ensure ``self.close`` gets called before GC
        atexit.register(self.close)

    @staticmethod
    def supports(uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        """
        Return if a given table is supported by the adapter.

        The discovery is done in 2 passes. First all adapters have their methods
        called with ``fast=True``. On the first pass adapters should implement
        a cheap method, without any network calls.

        If no adapter returns ``True`` a second pass is made with ``fast=False``
        using only adapters that returned ``None`` on the first pass. In this
        second pass adapters can perform network requests to get more
        information about the URI.

        The method receives the table URI, as well as the adapter connection
        arguments, eg::

            >>> from shillelagh.backends.apsw.db import connect
            >>> connection = connect(
            ...     ':memory:',
            ...     adapter_kwargs={"gsheetsapi": {"catalog":
            ...         {"table": "https://docs.google.com/spreadsheets/d/1/"}}},
            ... )

        This would call all adapters in order to find which one should handle
        the table ``table``. The Gsheets adapter would be called with::

            >>> from shillelagh.adapters.api.gsheets.adapter import GSheetsAPI
            >>> GSheetsAPI.supports("table", fast=True,  # first pass
            ...     catalog={"table": "https://docs.google.com/spreadsheets/d/1"})
            True

        """
        raise NotImplementedError("Subclasses must implement ``supports``")

    @staticmethod
    def parse_uri(uri: str) -> tuple[Any, ...]:
        """Parse table name, and return arguments to instantiate adapter."""
        raise NotImplementedError("Subclasses must implement ``parse_uri``")

    def get_metadata(self) -> dict[str, Any]:
        """Return any extra metadata about the table."""
        return {}

    def get_columns(self) -> dict[str, Field]:
        """
        Return the columns available in the table.

        This method is called for every query, so make sure it's cheap. For most
        (all?) tables this won't change, so you can store it in an instance
        attribute.
        """
        return dict(
            inspect.getmembers(self, lambda attribute: isinstance(attribute, Field)),
        )

    def get_cost(  # pylint: disable=unused-argument
        self,
        filtered_columns: list[tuple[str, Operator]],
        order: list[tuple[str, RequestedOrder]],
    ) -> float:
        """
        Estimate the query cost.

        The base adapter returns a fixed cost, and custom adapter can implement
        their own cost estimation.
        """
        return FIXED_COST

    def get_data(
        self,
        bounds: dict[str, Filter],
        order: list[tuple[str, RequestedOrder]],
        **kwargs: Any,
    ) -> Iterator[Row]:
        """
        Yield rows as adapter-specific types.

        This method expects rows to be in the storage format. Eg, for the CSV adapter
        datetime columns would be stored (and yielded) as strings. The ``get_rows``
        method will use the adapter fields to convert these values into native Python
        types (in this case, a proper ``datetime.datetime``).

        Missing values (NULLs) may be omitted from the dictionary; they will be
        replaced by ``None`` by the backend.
        """
        raise NotImplementedError("Subclasses must implement ``get_data``")

    def get_rows(
        self,
        bounds: dict[str, Filter],
        order: list[tuple[str, RequestedOrder]],
        **kwargs: Any,
    ) -> Iterator[Row]:
        """
        Yield rows as native Python types.
        """
        columns = self.get_columns()
        parsers = {column_name: field.parse for column_name, field in columns.items()}
        parsers["rowid"] = RowID().parse

        for row in self.get_data(bounds, order, **kwargs):
            yield {
                column_name: parsers[column_name](value)
                for column_name, value in row.items()
                if column_name in parsers
            }

    def insert_data(self, row: Row) -> int:
        """
        Insert a single row with adapter-specific types.

        The rows will be formatted according to the adapter fields. Eg, if an adapter
        represents timestamps as ISO strings, and timestamp values will be ISO strings.
        """
        raise NotSupportedError("Adapter does not support ``INSERT`` statements")

    def insert_row(self, row: Row) -> int:
        """
        Insert a single row with native Python types.

        The row types will be converted to the native adapter types, and passed to
        ``insert_data``.
        """
        columns = self.get_columns().copy()
        columns["rowid"] = RowID()
        row = {
            column_name: columns[column_name].format(value)
            for column_name, value in row.items()
        }
        return self.insert_data(row)

    def delete_data(self, row_id: int) -> None:
        """Delete a row from the table."""
        raise NotSupportedError("Adapter does not support ``DELETE`` statements")

    def delete_row(self, row_id: int) -> None:
        """
        Delete a row from the table.

        This method is identical to ``delete_data``, only here for symmetry.
        """
        return self.delete_data(row_id)

    def update_data(self, row_id: int, row: Row) -> None:
        """
        Update a single row with adapter-specific types.

        This method by default will call a delete followed by an insert.
        Adapters can implement their own more efficient methods.
        """
        try:
            self.delete_data(row_id)
            self.insert_data(row)
        except NotSupportedError as ex:
            raise NotSupportedError(
                "Adapter does not support ``UPDATE`` statements",
            ) from ex

    def update_row(self, row_id: int, row: Row) -> None:
        """
        Update a single row with native Python types.
        """
        columns = self.get_columns().copy()
        columns["rowid"] = RowID()
        row = {
            column_name: columns[column_name].format(value)
            for column_name, value in row.items()
        }
        self.update_data(row_id, row)

    def close(self) -> None:
        """
        Close the adapter.

        Adapters should use this method to perform any pending changes when the
        connection is closed.
        """

    def drop_table(self) -> None:
        """
        Drop a table.
        """



"""
An FDW.
"""

from collections import defaultdict
from collections.abc import Iterator
from typing import Any, DefaultDict, Optional, TypedDict

from multicorn import ForeignDataWrapper, Qual, SortKey

from shillelagh.adapters.registry import registry
from shillelagh.fields import Order
from shillelagh.filters import Operator
from shillelagh.lib import deserialize, get_bounds
from shillelagh.typing import RequestedOrder, Row

operator_map = {
    "=": Operator.EQ,
    ">": Operator.GT,
    "<": Operator.LT,
    ">=": Operator.GE,
    "<=": Operator.LE,
}


def get_all_bounds(quals: list[Qual]) -> DefaultDict[str, set[tuple[Operator, Any]]]:
    """
    Convert list of ``Qual`` into a set of operators for each column.
    """
    all_bounds: DefaultDict[str, set[tuple[Operator, Any]]] = defaultdict(set)
    for qual in quals:
        if operator := operator_map.get(qual.operator):
            all_bounds[qual.field_name].add((operator, qual.value))

    return all_bounds


class OptionsType(TypedDict):
    """
    Type for OPTIONS.
    """

    adapter: str
    args: str


class MulticornForeignDataWrapper(ForeignDataWrapper):
    """
    A FDW that dispatches queries to adapters.
    """

    def __init__(self, options: OptionsType, columns: dict[str, str]):
        super().__init__(options, columns)

        deserialized_args = deserialize(options["args"])
        self.adapter = registry.load(options["adapter"])(*deserialized_args)
        self.columns = self.adapter.get_columns()

    def execute(
        self,
        quals: list[Qual],
        columns: list[str],
        sortkeys: Optional[list[SortKey]] = None,
    ) -> Iterator[Row]:
        """
        Execute a query.
        """
        all_bounds = get_all_bounds(quals)
        bounds = get_bounds(self.columns, all_bounds)

        order: list[tuple[str, RequestedOrder]] = [
            (key.attname, Order.DESCENDING if key.is_reversed else Order.ASCENDING)
            for key in sortkeys or []
        ]

        kwargs = (
            {"requested_columns": columns}
            if self.adapter.supports_requested_columns
            else {}
        )

        return self.adapter.get_rows(bounds, order, **kwargs)

    def can_sort(self, sortkeys: list[SortKey]) -> list[SortKey]:
        """
        Return a list of sorts the adapter can perform.
        """

        def is_sortable(key: SortKey) -> bool:
            """
            Return if a given sort key can be enforced by the adapter.
            """
            if key.attname not in self.columns:
                return False

            order = self.columns[key.attname].order
            return (
                order == Order.ANY
                or (order == Order.ASCENDING and not key.is_reversed)
                or (order == Order.DESCENDING and key.is_reversed)
            )

        return [key for key in sortkeys if is_sortable(key)]

    def insert(self, values: Row) -> Row:
        rowid = self.adapter.insert_row(values)
        values["rowid"] = rowid
        return values

    def delete(self, oldvalues: Row) -> None:
        rowid = oldvalues["rowid"]
        self.adapter.delete_row(rowid)

    def update(self, oldvalues: Row, newvalues: Row) -> Row:
        rowid = newvalues["rowid"]
        self.adapter.update_row(rowid, newvalues)
        return newvalues

    @property
    def rowid_column(self):
        return "rowid"

    def get_rel_size(self, quals: list[Qual], columns: list[str]) -> tuple[int, int]:
        """
        Estimate query cost.
        """
        all_bounds = get_all_bounds(quals)
        filtered_columns = [
            (column, operator[0])
            for column, operators in all_bounds.items()
            for operator in operators
        ]

        # the adapter returns an arbitrary cost that takes in consideration filtering and
        # sorting; let's use that as an approximation for rows
        rows = int(self.adapter.get_cost(filtered_columns, []))

        # same assumption as the parent class
        row_width = len(columns) * 100

        return (rows, row_width)

    @classmethod
    def import_schema(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        cls,
        schema: str,
        srv_options: dict[str, str],
        options: dict[str, str],
        restriction_type: Optional[str],
        restricts: list[str],
    ):
        return []



"""
An adapter for fetching JSON data.
"""

# pylint: disable=invalid-name

import logging
from collections.abc import Iterator
from datetime import timedelta
from typing import Any, Optional, Union

import jsonpath
import prison
from yarl import URL

from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Field, Order
from shillelagh.filters import Filter
from shillelagh.lib import SimpleCostModel, analyze, flatten, get_session
from shillelagh.typing import Maybe, RequestedOrder, Row

_logger = logging.getLogger(__name__)

SUPPORTED_PROTOCOLS = {"http", "https"}
AVERAGE_NUMBER_OF_ROWS = 100
REQUEST_HEADERS_KEY = "_s_headers"
CACHE_EXPIRATION = timedelta(minutes=3)


class GenericJSONAPI(Adapter):
    """
    An adapter for fetching JSON data.
    """

    safe = True

    supports_limit = False
    supports_offset = False
    supports_requested_columns = True

    content_type = "application/json"
    default_path = "$[*]"
    cache_name = "generic_json_cache"

    @classmethod
    def supports(cls, uri: str, fast: bool = True, **kwargs: Any) -> Optional[bool]:
        parsed = URL(uri)
        if parsed.scheme not in SUPPORTED_PROTOCOLS:
            return False
        if fast:
            return Maybe

        if REQUEST_HEADERS_KEY in parsed.query:
            request_headers = prison.loads(parsed.query[REQUEST_HEADERS_KEY])
            parsed = parsed.with_query(
                {k: v for k, v in parsed.query.items() if k != REQUEST_HEADERS_KEY},
            )
        else:
            request_headers = kwargs.get("request_headers", {})

        cache_expiration = kwargs.get(
            "cache_expiration",
            CACHE_EXPIRATION.total_seconds(),
        )
        session = get_session(
            request_headers,
            cls.cache_name,
            timedelta(seconds=cache_expiration),
        )
        response = session.head(str(parsed))
        return cls.content_type in response.headers.get("content-type", "")

    @classmethod
    def parse_uri(
        cls,
        uri: str,
    ) -> Union[tuple[str, str], tuple[str, str, dict[str, str]]]:
        parsed = URL(uri)

        path = parsed.fragment or cls.default_path
        parsed = parsed.with_fragment("")

        if REQUEST_HEADERS_KEY in parsed.query:
            request_headers = prison.loads(parsed.query[REQUEST_HEADERS_KEY])
            parsed = parsed.with_query(
                {k: v for k, v in parsed.query.items() if k != REQUEST_HEADERS_KEY},
            )
            return str(parsed), path, request_headers

        return str(parsed), path

    def __init__(
        self,
        uri: str,
        path: Optional[str] = None,
        request_headers: Optional[dict[str, str]] = None,
        cache_expiration: float = CACHE_EXPIRATION.total_seconds(),
    ):
        super().__init__()

        self.uri = uri
        self.path = path or self.default_path

        self._session = get_session(
            request_headers or {},
            self.cache_name,
            timedelta(seconds=cache_expiration),
        )

        self._set_columns()

    def _set_columns(self) -> None:
        rows = list(self.get_data({}, []))
        column_names = list(rows[0].keys()) if rows else []

        _, order, types = analyze(iter(rows))

        self.columns = {
            column_name: types[column_name](
                filters=[],
                order=order.get(column_name, Order.NONE),
                exact=False,
            )
            for column_name in column_names
            if column_name != "rowid"
        }

    def get_columns(self) -> dict[str, Field]:
        return self.columns

    get_cost = SimpleCostModel(AVERAGE_NUMBER_OF_ROWS)

    def get_data(  # pylint: disable=unused-argument, too-many-arguments
        self,
        bounds: dict[str, Filter],
        order: list[tuple[str, RequestedOrder]],
        requested_columns: Optional[set[str]] = None,
        **kwargs: Any,
    ) -> Iterator[Row]:
        response = self._session.get(self.uri)
        payload = response.json()
        if not response.ok:
            raise ProgrammingError(f'Error: {payload["error"]["message"]}')

        for i, row in enumerate(jsonpath.findall(self.path, payload)):
            if isinstance(row, list):
                row = {f"col_{i}": value for i, value in enumerate(row)}
            elif isinstance(row, str):
                row = {"col_0": row}
            elif row is None:
                row = {}

            row = {
                k: v
                for k, v in row.items()
                if requested_columns is None or k in requested_columns
            }
            row["rowid"] = i
            _logger.debug(row)
            yield flatten(row)



"""
A SQLAlchemy dialect based on psycopg2 and multicorn2.
"""

# pylint: disable=protected-access, abstract-method

from typing import Any, Optional, cast

from psycopg2 import extensions
from sqlalchemy.dialects.postgresql.psycopg2 import PGDialect_psycopg2
from sqlalchemy.engine.url import URL
from sqlalchemy.pool.base import _ConnectionFairy

from shillelagh.adapters.base import Adapter
from shillelagh.backends.multicorn import db
from shillelagh.exceptions import ProgrammingError
from shillelagh.lib import find_adapter


class Multicorn2Dialect(PGDialect_psycopg2):
    """
    A SQLAlchemy dialect for Shillelagh based on psycopg2 and multicorn2.
    """

    name = "shillelagh"
    driver = "multicorn2"

    supports_statement_cache = True

    @classmethod
    def dbapi(cls):  # pylint: disable=method-hidden
        """
        Return the DB API module.
        """
        return db

    @classmethod
    def import_dbapi(cls):
        """
        New version of the ``dbapi`` method.
        """
        return db

    def __init__(
        self,
        adapters: Optional[list[str]] = None,
        adapter_kwargs: Optional[dict[str, dict[str, Any]]] = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._adapters = adapters
        self._adapter_kwargs = adapter_kwargs or {}

    def create_connect_args(
        self,
        url: URL,
    ) -> tuple[list[Any], dict[str, Any]]:
        args, kwargs = super().create_connect_args(url)
        kwargs.update(
            {
                "adapters": self._adapters,
                "adapter_kwargs": self._adapter_kwargs,
            },
        )
        return args, kwargs

    def has_table(
        self,
        connection: _ConnectionFairy,
        table_name: str,
        schema: Optional[str] = None,
        **kwargs: Any,
    ) -> bool:
        """
        Return true if a given table exists.
        """
        try:
            get_adapter_for_table_name(connection, table_name)
        except ProgrammingError:
            return bool(
                super().has_table(
                    connection,
                    table_name,
                    schema,
                    **kwargs,
                ),
            )
        return True


def get_adapter_for_table_name(
    connection: _ConnectionFairy,
    table_name: str,
) -> Adapter:
    """
    Return an adapter associated with a connection.

    This function instantiates the adapter responsible for a given table name,
    using the connection to properly pass any adapter kwargs.
    """
    raw_connection = cast(extensions.connection, connection.engine.raw_connection())
    cursor = raw_connection.cursor()
    adapter, args, kwargs = find_adapter(
        table_name,
        cursor._adapter_kwargs,
        cursor._adapters,
    )
    return adapter(*args, **kwargs)
